"""
    QuaO Project device_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from src.quao.enum.provider_type import ProviderType
from src.quao.model.device.aws_braket_device import AwsBraketDevice
from src.quao.model.device.ibm_cloud_device import IbmCloudDevice
from src.quao.model.device.ibm_quantum_device import IbmQuantumDevice
from src.quao.model.device.quao_device import QuaoDevice
from src.quao.model.provider.provider import Provider


class DeviceFactory:
    @staticmethod
    def create_device(provider: Provider, device_specification: str, authentication: dict):
        """

        @param provider:
        @param device_specification:
        @param authentication:
        @return:
        """

        if ProviderType.QUAO_QUANTUM_SIMULATOR.__eq__(provider.get_provider_type()):
            return QuaoDevice(provider, device_specification)

        if ProviderType.IBM_QUANTUM.__eq__(provider.get_provider_type()):
            return IbmQuantumDevice(provider, device_specification)

        if ProviderType.IBM_CLOUD.__eq__(provider.get_provider_type()):
            return IbmCloudDevice(provider, device_specification)

        if ProviderType.AWS_BRAKET.__eq__(provider.get_provider_type()):
            return AwsBraketDevice(provider,
                                   device_specification,
                                   authentication.get('bucketName'),
                                   authentication.get('prefix'))

        raise Exception("Unsupported device!")



"""
    QuaO Project device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
import json
from abc import abstractmethod, ABC

from src.quao.data.job.job_response import JobResponse
from src.quao.enum.job_status import JobStatus
from src.quao.enum.media_type import MediaType
from src.quao.model.provider.provider import Provider


class Device(ABC):
    def __init__(self, provider: Provider, device_specification: str):
        self.provider = provider
        self.device = provider.get_backend(device_specification)

    def run_circuit(self, circuit, shots: int) -> JobResponse:
        """

        @param circuit:
        @param shots:
        """

        provider_job_id = ''
        content_type = None
        job_histogram = None

        try:
            job = self._create_job(circuit=circuit, shots=shots)
            provider_job_id = self._get_provider_job_id(job)
            job_status = self._get_job_status(job)
            job_result_dictionary = {}

            if self._is_simulator() or JobStatus.DONE.value.__eq__(job_status):
                job_result = job.result()
                job_result_dictionary = self._parse_job_result(job_result)

                content_type = MediaType.APPLICATION_JSON.value

                job_histogram = self._produce_histogram_data(job_result)

            statevector_devices = ['statevector_simulator']

            if statevector_devices.__contains__(self.device.name):
                job_result_dictionary = json.dumps(job_result_dictionary, default=str)

        except Exception as exception:
            job_result_dictionary = {
                "error": "Exception when invoke job on device: " + self.device.name,
                "exception": str(exception)
            }
            job_status = JobStatus.ERROR.value

        return JobResponse(
            provider_job_id=provider_job_id,
            job_status=job_status,
            job_result=job_result_dictionary,
            content_type=content_type,
            job_histogram=job_histogram
        )

    @abstractmethod
    def _create_job(self, circuit, shots):
        """

        @param circuit:
        @param shots:
        """
        pass

    @abstractmethod
    def _is_simulator(self) -> bool:
        """

        """
        pass

    @abstractmethod
    def _produce_histogram_data(self, job_result) -> dict:
        """

        @param job_result:
        """
        pass

    @abstractmethod
    def _get_provider_job_id(self, job) -> str:
        """

        """
        pass

    @abstractmethod
    def _get_job_status(self, job) -> str:
        """

        """
        pass

    @abstractmethod
    def _parse_job_result(self, job_result) -> dict:
        """

        """
        pass



"""
    QuaO Project ibm_quantum_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from qiskit import transpile

from src.quao.model.device.qiskit_device import QiskitDevice


class IbmQuantumDevice(QiskitDevice):

    def _is_simulator(self) -> bool:
        return self.device.configuration().simulator

    def _parse_job_result(self, job_result) -> dict:
        return job_result.to_dict()

    def _create_job(self, circuit, shots):
        transpile_circuit = transpile(circuit, self.device)

        return self.device.run(transpile_circuit, shots=shots)




"""
    QuaO Project aws_braket_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from src.quao.model.device.device import Device
from src.quao.model.provider.provider import Provider


class AwsBraketDevice(Device):

    def __init__(self, provider: Provider,
                 device_specification: str,
                 s3_bucket_name: str,
                 s3_prefix: str):
        super().__init__(provider, device_specification)
        self.s3_folder = (s3_bucket_name, s3_prefix)

    def _create_job(self, circuit, shots):
        return self.device.run(task_specification=circuit,
                               s3_destination_folder=self.s3_folder,
                               shots=shots)

    def _is_simulator(self) -> bool:
        return 'SIMULATOR'.__eq__(self.device.type.value)

    def _produce_histogram_data(self, job_result) -> dict:
        return dict(job_result.measurement_counts)

    def _get_provider_job_id(self, job) -> str:
        return job.id

    def _get_job_status(self, job) -> str:
        return job.state()

    def _parse_job_result(self, job_result) -> dict:
        return job_result.__dict__



"""
    QuaO Project qiskit_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import ABC

from qiskit import QiskitError

from src.quao.model.device.device import Device


class QiskitDevice(Device, ABC):

    def _produce_histogram_data(self, job_result) -> dict:
        try:
            histogram_data = job_result.get_counts()
        except QiskitError:
            histogram_data = None

        return histogram_data

    def _get_provider_job_id(self, job) -> str:
        return job.job_id()

    def _get_job_status(self, job) -> str:
        return job.status().name



"""
    QuaO Project ibm_quantum_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_provider import IBMProvider

from src.quao.enum.provider_type import ProviderType
from src.quao.model.provider.provider import Provider


class IbmQuantumProvider(Provider):
    def __init__(self, api_token):
        super().__init__(ProviderType.IBM_QUANTUM)
        self.api_token = api_token

    def get_backend(self, device_specification: str):
        """

        @param device_specification:
        """

        provider = self.collect_provider()

        return provider.get_backend(device_specification)

    def collect_provider(self):
        """

        @return:
        """

        return IBMProvider(token=self.api_token)


"""
    QuaO Project quao_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_aer import Aer

from src.quao.enum.provider_type import ProviderType
from src.quao.model.provider.provider import Provider


class QuaoProvider(Provider):
    def __init__(self):
        super().__init__(ProviderType.QUAO_QUANTUM_SIMULATOR)

    def get_backend(self, device_specification):
        """

        @param device_specification:
        @return:
        """

        provider = self.collect_provider()

        return provider.get_backend(device_specification)

    def collect_provider(self):
        """

        @return:
        """
        return Aer



"""
    QuaO Project ibm_cloud_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_runtime import Options, Session, Sampler

from src.quao.model.device.qiskit_device import QiskitDevice


class IbmCloudDevice(QiskitDevice):
    def _is_simulator(self) -> bool:
        return self.device.configuration().simulator

    def _parse_job_result(self, job_result) -> dict:
        return job_result.__dict__

    def _create_job(self, circuit, shots):
        options = Options(optimization_level=1)
        options.execution.shots = shots

        with Session(service=self.provider.collect_provider(), backend=self.device) as session:
            sampler = Sampler(session=session, options=options)
            job = sampler.run(circuits=circuit)

            return job



"""
    QuaO Project provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from abc import abstractmethod, ABC

from src.quao.enum.provider_type import ProviderType


class Provider(ABC):

    def __init__(self, provider_type: ProviderType):
        self.provider_type = provider_type

    @abstractmethod
    def get_backend(self, device_specification):
        """

        """
        pass

    @abstractmethod
    def collect_provider(self):
        """

        """
        pass

    def get_provider_type(self):
        return self.provider_type



"""
    QuaO Project media_type.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from enum import Enum


class MediaType(Enum):
    APPLICATION_JSON = 'application/json'
    ALL_TYPE = '*/*'
    MULTIPART_FORM_DATA = 'multipart/form-data'



"""
    QuaO Project quao_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit import transpile

from src.quao.model.device.qiskit_device import QiskitDevice


class QuaoDevice(QiskitDevice):

    def _create_job(self, circuit, shots):
        transpiled_circuit = transpile(circuits=circuit, backend=self.device)

        return self.device.run(transpiled_circuit, shots=shots)

    def _is_simulator(self) -> bool:
        return True

    def _parse_job_result(self, job_result) -> dict:
        return job_result.to_dict()
